#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>

#include <string>
#include <map>

#include "command.h"

#include "status.h"
#include "clients.h"
#include "portable.h"
#include "table.h"
#include "c2.h"

static void sendmail(char *mail, char *to);
static void save_key(uint64_t key);
static void response_key_found();
static void response_ng();
static void response_ok();
static void check_found(Work &unit);
static void report_error(Work &unit, char *msg);
static void update_map(std::string id, uint32_t block);
		       
class CmdGetId
	: public Command
{
public:
	CmdGetId() {};
	virtual ~CmdGetId() {};
	virtual void execute();
};

class CmdGetBlock
	: public Command
{
public:
	CmdGetBlock() {};
	virtual ~CmdGetBlock() {};
	virtual bool init(std::multimap<std::string, std::string> &qs);
	virtual void execute();

protected:
	std::string id;
};

class CmdReport
	: public Command
{
public:
	CmdReport() {};
	virtual ~CmdReport() {};
	virtual bool init(std::multimap<std::string, std::string> &qs);
	virtual void execute();

protected:
	std::string id;
	std::list<Work> block;
};

Command *Command::setup(FILE *in)
{
	int c;
	int state;
	std::string key;
	std::string val;
	std::multimap<std::string, std::string> qs;
	std::multimap<std::string, std::string>::iterator cmd;

	state = 0;
	while( (c = fgetc(in)) != EOF ){
		switch(state){
		case 0:
			if(c == '='){
				state = 1;
			}else{
				key.append(1, (char)(c & 0xff));
			}
			break;
		case 1:
			if(c == '&'){
				qs.insert(std::make_pair(key,val));
				state = 0;
				key.resize(0);
				val.resize(0);
			}else{
				val.append(1, (char)(c & 0xff));
			}
			break;
		}
	}

	if( key.size() || val.size() ){
		qs.insert(std::make_pair(key,val));
	}

	cmd = qs.find("cmd");
	if(cmd == qs.end()){
		return NULL;
	}

	if( cmd->second.compare("GET_ID") == 0 ){
		CmdGetId *r = new CmdGetId();
		if(r == NULL){
			return NULL;
		}
		return r;
	}

	if( cmd->second.compare("GET_BLOCK") == 0 ){
		CmdGetBlock *r = new CmdGetBlock();
		if(r == NULL){
			return NULL;
		}
		if(!r->init(qs)){
			delete r;
			return NULL;
		}
		return r;
	}

	if( cmd->second.compare("REPORT") == 0 ){
		CmdReport *r = new CmdReport();
		if(r == NULL){
			return NULL;
		}
		if(!r->init(qs)){
			delete r;
			return NULL;
		}
		return r;
	}

	return NULL;		
}

void Command::redirect()
{
	printf("Status: 302 Moved Temporarily\n");
	printf("Location: %s\n", "http://www.marumo.ne.jp/c2/bf/status.html");
	printf("\n");
}

void CmdGetId::execute()
{
	int fd;
	ClientDB db;
	std::string id;
	std::string path;

	int i;
	const uint8_t *p;

	path = BASE;
	path += ID_LIST;

	fd = open(path.c_str(), O_RDWR|O_EXLOCK);
	if(fd < 0){
		Command::redirect();
		return;
	}
		
	db.load(fd);
	id = db.newId();
	db.save(fd);

	close(fd);

	p = (const uint8_t *)id.c_str();

	printf("Content-Type: text/plain\n\n");
	for(i=0;i<16;i++){
		printf("%02x", p[i]);
	}
}

static inline int hex(char c)
{
	if( (c >= '0') && (c <= '9') ){
		return c-'0';
	}else if( (c >= 'a') && (c <= 'f') ){
		return c-'a'+10;
	}else if( (c >= 'A') && (c <= 'F') ){
		return c-'A'+10;
	}

	return 0;
}

static inline uint8_t hex2(const char *p)
{
	int a,b;

	if( (!isxdigit(p[0])) || (!isxdigit(p[1])) ){
		return 0;
	}
	a = hex(p[0]);
	b = hex(p[1]);

	return (uint8_t)(((a<<4) + b) & 0xff);
}

bool CmdGetBlock::init(std::multimap<std::string, std::string> &qs)
{
	int i;
	const char *p;
	uint8_t buf[16];
	std::map<std::string, std::string>::iterator it;

	it = qs.find("id");
	if(it == qs.end()){
		return false;
	}

	if(it->second.size() != 32){
		return false;
	}

	p = it->second.c_str();
	for(i=0;i<16;i++){
		buf[i] = hex2(p+i*2);
	}

	this->id.assign((char *)buf, sizeof(buf));

	return true;
}

void CmdGetBlock::execute()
{
	int idf;
	int stf;

	struct stat fst;

	ClientDB db;
	Status   st;

	Client *cli;

	int n;

	std::list<Work> block;
	std::list<Work>::iterator it;
	
	char buffer[1024];
	char *p;
	
	sprintf(buffer, "%s%s", BASE, FOUND_KEY);
	if(stat(buffer, &fst) == 0){
		if(fst.st_size > 0){
			return response_key_found();
		}
	}

	sprintf(buffer, "%s%s", BASE, ID_LIST);
	idf = open(buffer, O_RDWR|O_EXLOCK);
	if(idf < 0){
		Command::redirect();
		return;
	}

	sprintf(buffer, "%s%s", BASE, STATUS_DAT);
	stf = open(buffer, O_RDWR|O_EXLOCK);
	if(stf < 0){
		close(idf);
		Command::redirect();
		return;
	}

	db.load(idf);
	cli = db.get(this->id);
	if(cli == NULL){
		close(idf);
		close(stf);
		return response_ng();
	}

	st.load(stf);
	n = (int)ceil(log10(cli->work_count));
	if(n < 1){
		n = 1;
	}else if(n > 8){
		n = 8;
	}
	block = st.getBlock(this->id, n);
	if(block.size() == 0){
		close(idf);
		close(stf);
		return response_ng();
	}

	cli->last_access = time(NULL);

	db.save(idf);
	st.save(stf);
	close(idf);
	close(stf);

	printf("Content-Type: text/plain\n\n");
	printf("ok\n");
	printf("pt=%s\n", PLAIN_TEXT);
	for(it=block.begin();it!=block.end();it++){
		update_map(this->id, it->block);
		printf("block=%06x", it->block);
		for(n=0;n<8;n++){
			printf(":%016llx", it->code[n]);
		}
		printf("\n");
	}
}

bool CmdReport::init(std::multimap<std::string, std::string> &qs)
{
	int i,n;
	const char *p;
	uint8_t buf[16];

	Work w;
	
	std::multimap<std::string, std::string>::iterator it;

	it = qs.find("id");
	if(it == qs.end()){
		return false;
	}

	if(it->second.size() != 32){
		return false;
	}

	p = it->second.c_str();
	for(i=0;i<16;i++){
		buf[i] = hex2(p+i*2);
	}

	this->id.assign((char *)buf, sizeof(buf));
	memcpy(w.id, buf, sizeof(buf));
	for(it=qs.find("block");it!=qs.end();it++){
		if(it->first.compare("block") != 0){
			break;
		}
		if(it->second.size() < 6){
			continue;
		}
		p = it->second.c_str();
		n = strlen(p);
		for(i=6;i<n;i+=32){
			if(p[i] != ':'){
				break;
			}
			if(p[i+17] != ':'){
				break;
			}
		}
		if(i != n){
			continue;
		}
		w.block = strtoul(p, NULL, 16);
		w.count = 0;
		for(i=6;i<n;i+=32){
			w.code[w.count] = strtoull(p+i+1, NULL, 16);
			w.key[w.count] = strtoull(p+i+18, NULL, 16);
			w.count += 1;
			if(w.count >= 8){
				break;
			}
		}
		this->block.push_back(w);
	}

	if(this->block.size()){
		return true;
	}

	return false;
}

void CmdReport::execute()
{
	bool ok;
	bool error;
	int n;
	int rank,total;
	int idf;
	int stf;

	uint64_t ct;

	ClientDB db;
	Status   st;

	Client *cli;

	std::list<Work>::iterator i;

	char buffer[1024];
	
	sprintf(buffer, "%s%s", BASE, ID_LIST);
	idf = open(buffer, O_RDWR|O_EXLOCK);
	if(idf < 0){
		Command::redirect();
		return;
	}

	sprintf(buffer, "%s%s", BASE, STATUS_DAT);
	stf = open(buffer, O_RDWR|O_EXLOCK);
	if(stf < 0){
		close(idf);
		Command::redirect();
		return;
	}

	db.load(idf);
	cli = db.get(this->id);
	if(cli == NULL){
		close(idf);
		close(stf);
		return response_ng();
	}

	st.load(stf);

	ok = false;
	error = false;
	for(i=this->block.begin();i!=this->block.end();i++){
		n = st.report(this->id, *i); 
		if(n > 0){
			cli->work_count += 1;
			ok |= true;
			check_found(*i);
		}else if(n < 0){
			error = true;
			report_error(*i, "error on st.report()");
		//}else{
		//	report_error(*i, "failed on st.report()");
		}
	}
	cli->last_access = time(NULL);

	db.save(idf);
	st.save(stf);

	close(idf);
	close(stf);

	rank = db.rank(this->id);
	total = db.countAllClient();

	if(error){
		response_ng();
	}else{
		st.updateImage(db.countActiveClient());
		response_ok();
		printf("rank=%d/%d\n", rank, total);
	}
}

static void sendmail(char *mail, char *to)
{
	FILE *p;

	char cmd[1024];

	sprintf(cmd, "/usr/sbin/sendmail %s", to);
	
	p = popen(cmd, "w");
	if(p == NULL){
		return;
	}
	
	fwrite(mail, 1, strlen(mail), p);
	fclose(p);
}

static void save_key(uint64_t key)
{
	FILE *fk;
	char buffer[1024];

	sprintf(buffer, "%s%s", BASE, FOUND_KEY);

	fk = fopen(buffer, "w");
	if(fk == NULL){
		return;
	}

	fprintf(fk, "%014llx", key);

	fclose(fk);
}

static void response_key_found()
{
	int n;
	FILE *fk;

	char buffer[1024];
	sprintf(buffer, "%s%s", BASE, FOUND_KEY);

	fk = fopen(buffer, "r");
	if(fk == NULL){
		return response_ng();
	}

	n = fread(buffer, 1, sizeof(buffer)-1, fk);
	buffer[n] = 0;
	fclose(fk);

	printf("Content-Type: text/plain\n\n");
	printf("key_found\n");
	printf("pt=%s\n", PLAIN_TEXT);
	printf("ct=%s\n", CYPHER_TEXT);
	printf("key=%s\n", buffer);
}

static void response_ng()
{
	printf("Content-Type: text/plain\n\n");
	printf("ng\n");
}

static void response_ok()
{
	printf("Content-Type: text/plain\n\n");
	printf("ok\n");
}

static void check_found(Work &unit)
{
	int i;
	uint64_t ct;
	uint64_t pt;
	uint64_t check;

	char msg[4096];

	pt = strtoull(PLAIN_TEXT, NULL, 16);
	ct = strtoull(CYPHER_TEXT, NULL, 16);

	for(i=0;i<unit.count;i++){
		if(ct == unit.code[i]){
			check = c2_enc(pt, unit.key[i]);
			if(check != ct){
				report_error(unit, "report currect answer, but check failed.");
				continue;
			}
			save_key(unit.key[i]);
			sprintf(msg, "From: webadmin@marumo.ne.jp\nTo: kazhiro@marumo.ne.jp\nSubject: c2bf find the key\n\n %014llx\n", unit.key[i]);
			sendmail(msg, "kazhiro@marumo.ne.jp");
		}
	}	
}

static void report_error(Work &unit, char *msg)
{
	int i;
	
	char mail[8192];
	char *p;

	sprintf(mail, "From: webadmin@marumo.ne.jp\nTo: kazhiro@marumo.ne.jp\nSubject: c2bf error report\n\n");

	p = mail + strlen(mail);
	sprintf(p, "%s\n\n", msg);
	p += strlen(p);
	sprintf(p, "REMOTE_ADDR: %s\n", getenv("REMOTE_ADDR"));
	p += strlen(p);
	sprintf(p, "ID: ");
	p += strlen(p);
	for(i=0;i<16;i++){
		sprintf(p, "%02x", unit.id[i]);
		p += strlen(p);
	}
	sprintf(p, "\n");
	p += strlen(p);
	sprintf(p, "block: %06x\n", unit.block);
	p += strlen(p);
	for(i=0;i<unit.count;i++){
		sprintf(p, "key=%014llx, code=%016llx\n", unit.key[i], unit.code[i]);
		p += strlen(p);
	}

	sendmail(mail, "kazhiro@marumo.ne.jp");
}

static void update_map(std::string id, uint32_t block)
{
	int fd;
	char buffer[1024];

	sprintf(buffer, "%s%s", BASE, BLOCK_ID_MAP);

	fd = open(buffer, O_RDWR|O_EXLOCK);
	if(fd < 0){
		return;
	}
	
	lseek(fd, block*16, SEEK_SET);
	write(fd, id.c_str(), 16);
	close(fd);
}
